#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import json
import shutil
import numpy as np
import pandas as pd
from datasets.arrow_dataset import Batch
from transformers import TextClassificationPipeline
from transformers.trainer_utils import EvalPrediction
from transformers.models.distilbert.modeling_distilbert import PreTrainedModel
from transformers.tokenization_utils_base import BatchEncoding, PreTrainedTokenizerBase

from {{package_name}} import utils
from {{package_name}}.models_training.model_huggingface import ModelHuggingFace

# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class ModelHuggingFaceTests(unittest.TestCase):
    '''Main class to test model_huggingface'''

    def setUp(self):
        '''Setup fonction -> we create a mock embedding'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    def test01_model_huggingface_init(self):
        '''Test of the initialization of {{package_name}}.models_training.model_huggingface.ModelHuggingFace'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Init., test all parameters
        model = ModelHuggingFace(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.model, None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        #
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8)
        self.assertEqual(model.batch_size, 8)
        remove_dir(model_dir)

        #
        model = ModelHuggingFace(model_dir=model_dir, epochs=42)
        self.assertEqual(model.epochs, 42)
        remove_dir(model_dir)

        #
        model = ModelHuggingFace(model_dir=model_dir, validation_split=0.3)
        self.assertEqual(model.validation_split, 0.3)
        remove_dir(model_dir)

        #
        model = ModelHuggingFace(model_dir=model_dir, patience=10)
        self.assertEqual(model.patience, 10)
        remove_dir(model_dir)

        #
        model = ModelHuggingFace(model_dir=model_dir, model_max_length=10)
        self.assertEqual(model.model_max_length, 10)
        remove_dir(model_dir)

        # Can't be tested as this would try to load a transformer called 'toto'
        # We could patch it, but w/e
        # model = ModelHuggingFace(model_dir=model_dir, transformer_name='toto')
        # self.assertEqual(model.transformer_name, 'toto')
        # remove_dir(model_dir)

        # transformer_params must accept anything !
        model = ModelHuggingFace(model_dir=model_dir, transformer_params={'toto': 5})
        self.assertEqual(model.transformer_params, {'toto': 5})
        remove_dir(model_dir)

        # trainer_params
        model = ModelHuggingFace(model_dir=model_dir, trainer_params={'toto': 5})
        self.assertEqual(model.trainer_params, {'toto': 5})
        remove_dir(model_dir)

    def test02_model_huggingface_fit(self):
        '''Test of the method fit of {{package_name}}.models_training.model_huggingface.ModelHuggingFace'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"] * 100)
        x_valid = np.array(["cela est un test", "ni cela", "non plus", "ici test", "là, rien de rien!"] * 100)
        y_train_mono = np.array([0, 1, 0, 1, 2] * 100)
        y_valid_mono = y_train_mono.copy()
        y_valid_mono_missing = y_train_mono.copy()
        y_valid_mono_missing[y_valid_mono_missing == 2] = 0
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0] * 100, 'test2': [1, 0, 0, 0, 0] * 100, 'test3': [0, 0, 0, 1, 0] * 100})
        y_valid_multi = y_train_multi.copy()
        x_train_long = np.array(["cela est un test " * 1000, "là, rien! " * 1000] * 10)
        y_train_long = np.array([0, 1] * 10)

        # Mono-label
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_mono, x_valid=None, y_valid=None, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), [0, 1, 2])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)

        # with valid data
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_mono, x_valid=x_valid, y_valid=y_valid_mono, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), [0, 1, 2])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)

        # no shuffle
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_mono, x_valid=x_valid, y_valid=y_valid_mono, with_shuffle=False)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), [0, 1, 2])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)

        # Validation with y_train & y_valid of shape 2
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, np.expand_dims(y_train_mono, 1), x_valid=x_valid, y_valid=np.expand_dims(y_valid_mono, 1), with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), [0, 1, 2])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)

        # Missing targets in y_valid
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_mono, x_valid=x_valid, y_valid=y_valid_mono_missing, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), [0, 1, 2])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)

        # Multi-labels
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_multi, x_valid=None, y_valid=None, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), ['test1', 'test2', 'test3'])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)

        # with valid
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_multi, x_valid=x_valid, y_valid=y_valid_multi, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), ['test1', 'test2', 'test3'])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)

        # No shuffle
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_multi, x_valid=x_valid, y_valid=y_valid_multi, with_shuffle=False)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), ['test1', 'test2', 'test3'])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)


        ###########
        # Test continue training

        # Test mono-label nominal case
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_mono, x_valid=None, y_valid=None, with_shuffle=True)
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # Second fit
        y_train_different_order = np.array([1, 0, 0, 1, 2] * 100)
        model.fit(x_train[:50], y_train_different_order[:50], x_valid=None, y_valid=None, with_shuffle=True)
        # We do not save on purpose
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 2)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        model_dir_2 = model.model_dir
        self.assertNotEqual(model_dir, model_dir_2)
        # third fit
        model.fit(x_train[50:], y_train_mono[50:], x_valid=None, y_valid=None, with_shuffle=True)
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 3)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_2.json')))
        model_dir_3 = model.model_dir
        self.assertNotEqual(model_dir_2, model_dir_3)
        # Fourth fit
        model.fit(x_train[50:], y_train_mono[50:], x_valid=None, y_valid=None, with_shuffle=True)
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 4)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_2.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_3.json')))
        model_dir_4 = model.model_dir
        self.assertNotEqual(model_dir_3, model_dir_4)
        remove_dir(model_dir)
        remove_dir(model_dir_2)
        remove_dir(model_dir_3)
        remove_dir(model_dir_4)

        # Test data errors mono-label
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_mono, x_valid=None, y_valid=None, with_shuffle=True)
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # Second fit
        y_train_mono_fake = np.array([3, 1, 0, 1, 2] * 100)
        with self.assertRaises(ValueError):
            model.fit(x_train[:50], y_train_mono_fake[:50], x_valid=None, y_valid=None, with_shuffle=True)
        remove_dir(model_dir)

        # Test multi-labels nominal case
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_multi, x_valid=None, y_valid=None, with_shuffle=True)
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # Second fit
        model.fit(x_train[:50], y_train_multi[:50], x_valid=None, y_valid=None, with_shuffle=True)
        # We do not save on purpose
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 2)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        model_dir_2 = model.model_dir
        self.assertNotEqual(model_dir, model_dir_2)
        # third fit
        model.fit(x_train[50:], y_train_multi[50:], x_valid=None, y_valid=None, with_shuffle=True)
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 3)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        model_dir_3 = model.model_dir
        self.assertNotEqual(model_dir_2, model_dir_3)
        # Fourth fit
        model.fit(x_train[50:], y_train_multi[50:], x_valid=None, y_valid=None, with_shuffle=True)
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 4)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_2.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_3.json')))
        model_dir_4 = model.model_dir
        self.assertNotEqual(model_dir_3, model_dir_4)
        remove_dir(model_dir)
        remove_dir(model_dir_2)
        remove_dir(model_dir_3)
        remove_dir(model_dir_4)


        ###########
        # Misc
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train_long, y_train_long, x_valid=None, y_valid=None, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), [0, 1])
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        remove_dir(model_dir)

        # Test data errors multi-labels
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(x_train, y_train_multi, x_valid=None, y_valid=None, with_shuffle=True)
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # Second fit
        y_train_multi_fake = pd.DataFrame({'test3': [0, 0, 0, 1, 0] * 100, 'test2': [1, 0, 0, 0, 0] * 100, 'test1': [0, 0, 0, 1, 0] * 100})
        with self.assertRaises(ValueError):
            model.fit(x_train[:50], y_train_multi_fake[:50], x_valid=None, y_valid=None, with_shuffle=True)
        remove_dir(model_dir)

    def test03_model_huggingface_predict(self):
        '''Test of the method predict of {{package_name}}.models_training.model_huggingface.ModelHuggingFace'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"] * 100)
        x_valid = np.array(["cela est un test", "ni cela", "non plus", "ici test", "là, rien de rien!"] * 100)
        x_valid_long = np.array(["cela est un test " * 1000] * 10)
        y_train_mono = np.array([0, 1, 0, 1, 2] * 100)
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0] * 100, 'test2': [1, 0, 0, 0, 0] * 100, 'test3': [0, 0, 0, 1, 0] * 100})
        cols = ['test1', 'test2', 'test3']

        # Mono-label
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        model.fit(x_train, y_train_mono)
        preds = model.predict(x_valid, return_proba=False)
        self.assertEqual(preds.shape, (len(x_valid),))
        preds = model.predict('test', return_proba=False)
        self.assertEqual(preds, model.predict(['test'], return_proba=False)[0])
        proba = model.predict(x_valid, return_proba=True)
        self.assertEqual(proba.shape, (len(x_valid), 3))
        proba = model.predict('test', return_proba=True)
        self.assertEqual([elem for elem in proba], [elem for elem in model.predict(['test'], return_proba=True)[0]])
        # Predict long sentence (just check no errors)
        model.predict(x_valid_long)
        remove_dir(model_dir)

        # Multi-labels
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True)
        model.fit(x_train, y_train_multi)
        preds = model.predict(x_valid, return_proba=False)
        self.assertEqual(preds.shape, (len(x_valid), len(cols)))
        preds = model.predict('test', return_proba=False)
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict(['test'], return_proba=False)[0]])
        proba = model.predict(x_valid, return_proba=True)
        self.assertEqual(proba.shape, (len(x_valid), len(cols)))
        proba = model.predict('test', return_proba=True)
        self.assertEqual([elem for elem in proba], [elem for elem in model.predict(['test'], return_proba=True)[0]])
        # Predict long sentence (just check no errors)
        model.predict(x_valid_long)
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
            model.predict('test')
        remove_dir(model_dir)

    def test04_model_huggingface_predict_proba(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace.predict_proba'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"] * 100)
        x_test = np.array(["cela est un test", "ni cela", "non plus", "ici test", "là, rien de rien!"] * 100)
        x_test_long = np.array(["cela est un test " * 1000] * 10)
        y_train_mono = np.array([0, 1, 0, 1, 2] * 100)
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0] * 100, 'test2': [1, 0, 0, 0, 0] * 100, 'test3': [0, 0, 0, 1, 0] * 100})
        cols = ['test1', 'test2', 'test3']

        # Mono-label
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        model.fit(x_train, y_train_mono)
        preds = model.predict_proba(x_test)
        self.assertEqual(preds.shape, (len(x_test), 3))
        preds = model.predict_proba('test')
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict_proba(['test'])[0]])
        # Predict long sentence (just check no errors)
        model.predict_proba(x_test_long)
        remove_dir(model_dir)


        # Multi-labels
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True)
        model.fit(x_train, y_train_multi[cols])
        preds = model.predict_proba(x_test)
        self.assertEqual(preds.shape, (len(x_test), len(cols)))
        preds = model.predict_proba('test')
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict_proba(['test'])[0]])
        # Predict long sentence (just check no errors)
        model.predict_proba(x_test_long)
        remove_dir(model_dir)


        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
            model.predict_proba('test')
        remove_dir(model_dir)

    def test05_model_huggingface_prepare_x_train(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace._prepare_x_train'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        x_train = ['test titi toto', 'toto', 'titi test test toto']
        y_train_dummies = [[0, 0, 1], [1, 0, 0], [0, 1, 0]]

        # Nominal case
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        x_train_prepared = model._prepare_x_train(x_train, y_train_dummies)
        # We can't easily test the results, too many dependences
        self.assertEqual(x_train_prepared.shape[0], len(x_train))  # Same nb of lines
        self.assertTrue('text' in x_train_prepared.features)
        self.assertTrue('label' in x_train_prepared.features)
        self.assertTrue('input_ids' in x_train_prepared.features)
        self.assertTrue('attention_mask' in x_train_prepared.features)
        remove_dir(model_dir)

    def test06_model_huggingface_prepare_x_valid(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace._prepare_x_valid'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        x_valid = ['test titi toto', 'toto', 'titi test test toto']
        y_valid_dummies = [[0, 0, 1], [1, 0, 0], [0, 1, 0]]

        # Nominal case
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        x_valid_prepared = model._prepare_x_valid(x_valid, y_valid_dummies)
        # We can't easily test the results, too many dependences
        self.assertEqual(x_valid_prepared.shape[0], len(x_valid))  # Same nb of lines
        self.assertTrue('text' in x_valid_prepared.features)
        self.assertTrue('label' in x_valid_prepared.features)
        self.assertTrue('input_ids' in x_valid_prepared.features)
        self.assertTrue('attention_mask' in x_valid_prepared.features)
        remove_dir(model_dir)

    def test07_model_huggingface_prepare_x_test(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace._prepare_x_test'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        x_test = ['test titi toto', 'toto', 'titi test test toto']

        # Nominal case - tokenizer not set
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        # TMP FIX: init. tokenizer
        model.tokenizer = model._get_tokenizer()
        x_test_prepared = model._prepare_x_test(x_test)
        # We can't easily test the results, too many dependences
        self.assertEqual(x_test_prepared.shape[0], len(x_test))  # Same nb of lines
        self.assertTrue('text' in x_test_prepared.features)
        self.assertTrue('label' not in x_test_prepared.features)
        self.assertTrue('input_ids' in x_test_prepared.features)
        self.assertTrue('attention_mask' in x_test_prepared.features)
        remove_dir(model_dir)

    def test08_model_huggingface_tokenize_function(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace._tokenize_function'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        batch = Batch(data={'text': 'titi test test toto', 'label': 1})

        # Nominal case - tokenizer not set
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        # TMP FIX: init. tokenizer
        model.tokenizer = model._get_tokenizer()
        encoded_batch = model._tokenize_function(batch)
        # We can't easily test the results, too many dependences
        self.assertTrue(type(encoded_batch) == BatchEncoding)
        self.assertTrue('input_ids' in encoded_batch.keys())
        self.assertTrue('attention_mask' in encoded_batch.keys())
        remove_dir(model_dir)

    def test09_model_huggingface_get_model(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace._get_model'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)

        # Nominal case
        model.list_classes = ['a', 'b']  # We force the creation of a list of classes
        model_res = model._get_model()
        self.assertTrue(isinstance(model_res, PreTrainedModel))

        # Clean
        remove_dir(model_dir)

    def test10_model_huggingface_get_tokenizer(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace._get_tokenizer'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)

        # Nominal case
        tokenizer = model._get_tokenizer()
        self.assertTrue(isinstance(tokenizer, PreTrainedTokenizerBase))

        # With model_path
        # TODO

        # Clean
        remove_dir(model_dir)

    def test11_model_huggingface_compute_metrics_mono_label(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace._compute_metrics_mono_label'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)

        # Data to compute
        logits = np.array([[0.2, 0.8], [0.51, 0.49], [0.3, 0.7], [0.1, 0.9], [0.8, 0.2]])
        label_ids = np.array([1, 0, 0, 1, 0])
        weight_0 = 3 / 5
        weight_1 = 2 / 5
        eval_pred = EvalPrediction(logits, label_ids)

        # Nominal case
        metrics = model._compute_metrics_mono_label(eval_pred)
        self.assertTrue('accuracy' in metrics.keys())
        self.assertAlmostEqual(metrics['accuracy'], 4/5)
        self.assertTrue('weighted_precision' in metrics.keys())
        precision_0 = 2 / 2
        precision_1 = 2 / 3
        precision = (precision_0 * weight_0) + (precision_1 * weight_1)  # weighted precision
        self.assertAlmostEqual(metrics['weighted_precision'], precision)
        self.assertTrue('weighted_recall' in metrics.keys())
        recall_0 = 2 / 3
        recall_1 = 2 / 2
        recall = (recall_0 * weight_0) + (recall_1 * weight_1)  # weighted recall
        self.assertAlmostEqual(metrics['weighted_recall'], recall)
        self.assertTrue('weighted_f1' in metrics.keys())
        f1_0 =  2 * precision_0 * recall_0 / (precision_0 + recall_0)
        f1_1 =  2 * precision_1 * recall_1 / (precision_1 + recall_1)
        f1 = (f1_0 * weight_0) + (f1_1 * weight_1)  # weighted f1
        self.assertAlmostEqual(metrics['weighted_f1'], f1)

        # Clean
        remove_dir(model_dir)

    def test12_model_huggingface_compute_metrics_multi_label(self):
        '''Test of {{package_name}}.models_training.model_huggingface.ModelHuggingFace._compute_metrics_multi_label'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)

        # Data to compute
        # proba > 0.5 if logit > 0
        logits = np.array([[-1, 1], [0.1, -0.1], [-1, 1], [-1, -1], [1, 1]])
        # predictions = np.array([[0, 1], [1, 0], [0, 1], [0, 0], [1, 1]])
        label_ids = np.array([[0, 1], [1, 1], [1, 1], [0, 0], [0, 1]])
        weight_0 = 2 / 5
        weight_1 = 4 / 5
        total = weight_0 + weight_1
        weight_0 = weight_0 / total
        weight_1 = weight_1 / total
        eval_pred = EvalPrediction(logits, label_ids)

        # Nominal case
        metrics = model._compute_metrics_multi_label(eval_pred)
        self.assertTrue('accuracy' in metrics.keys())
        self.assertAlmostEqual(metrics['accuracy'], 2/5)
        self.assertTrue('weighted_precision' in metrics.keys())
        precision_0 = 1 / 2
        precision_1 = 3 / 3
        precision = (precision_0 * weight_0) + (precision_1 * weight_1)  # weighted precision
        self.assertAlmostEqual(metrics['weighted_precision'], precision)
        self.assertTrue('weighted_recall' in metrics.keys())
        recall_0 = 1 / 2
        recall_1 = 3 / 4
        recall = (recall_0 * weight_0) + (recall_1 * weight_1)  # weighted recall
        self.assertAlmostEqual(metrics['weighted_recall'], recall)
        self.assertTrue('weighted_f1' in metrics.keys())
        f1_0 =  2 * precision_0 * recall_0 / (precision_0 + recall_0)
        f1_1 =  2 * precision_1 * recall_1 / (precision_1 + recall_1)
        f1 = (f1_0 * weight_0) + (f1_1 * weight_1)  # weighted f1
        self.assertAlmostEqual(metrics['weighted_f1'], f1)

        # Clean
        remove_dir(model_dir)

    def test13_model_huggingface_save(self):
        '''Test of the method save of {{package_name}}.models_training.model_huggingface.ModelHuggingFace'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelHuggingFace(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False)
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_model', 'pytorch_model.bin')))
        # self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'hf_tokenizer', 'tokenizer.json'))) -> no model trained
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'model_huggingface.pkl')))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        self.assertTrue('x_col' in configs.keys())
        self.assertTrue('y_col' in configs.keys())
        self.assertTrue('multi_label' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'huggingface')
        self.assertTrue('batch_size' in configs.keys())
        self.assertTrue('epochs' in configs.keys())
        self.assertTrue('validation_split' in configs.keys())
        self.assertTrue('transformer_name' in configs.keys())
        self.assertTrue('transformer_params' in configs.keys())
        self.assertTrue('trainer_params' in configs.keys())
        self.assertTrue('_get_model' in configs.keys())
        self.assertTrue('_get_tokenizer' in configs.keys())

        # Clean
        remove_dir(model_dir)

    def test14_model_huggingface_init_new_instance_from_configs(self):
        '''Test of the method _init_new_instance_from_configs of {{package_name}}.models_training.models_tensorflow.model_huggingface.ModelHuggingFace'''
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelHuggingFace(model_dir=model_dir)
        model.save(json_data={'test': 8})
        configs = model.load_configs(model_dir=model_dir)

        new_model = ModelHuggingFace._init_new_instance_from_configs(configs=configs)
        self.assertTrue(isinstance(new_model, ModelHuggingFace))
        self.assertEqual(new_model.nb_fit, 0)
        self.assertFalse(new_model.trained)
        for attribute in ['x_col', 'y_col', 'list_classes', 'dict_classes', 'multi_label', 'level_save', 'batch_size', 'epochs', 'patience',
                          'transformer_name', 'model_max_length']:
            self.assertEqual(getattr(model, attribute), getattr(new_model, attribute))
        for attribute in ['validation_split']:
            self.assertAlmostEqual(getattr(model, attribute), getattr(new_model, attribute))
        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

        # Check by changing some attributes
        model = ModelHuggingFace(model_dir=model_dir)
        model.nb_fit = 2
        model.trained = True
        model.x_col = 'coucou'
        model.y_col = 'coucou_2'
        model.list_classes = ['class_1', 'class_2', 'class_3']
        model.dict_classes = {0: 'class_1', 1: 'class_2', 2: 'class_3'}
        model.multi_label = True
        model.level_save = 'MEDIUM'
        model.batch_size = 13
        model.epochs = 42
        model.validation_split = 0.3
        model.patience = 15
        model.transformer_name = 'coucou'
        model.model_max_length = 12
        model.trainer_params = {
                'weight_decay': 0.0,
                'evaluation_strategy': 'epoch',
                'load_best_model_at_end': True
            }
        model.transformer_params = {
                'weight_decay': 0.0,
                'evaluation_strategy': 'epoch',
                'load_best_model_at_end': True
            }
        model.save(json_data={'test': 8})
        configs = model.load_configs(model_dir=model_dir)

        new_model = ModelHuggingFace._init_new_instance_from_configs(configs=configs)
        self.assertTrue(isinstance(new_model, ModelHuggingFace))
        self.assertEqual(new_model.nb_fit, 2)
        self.assertTrue(new_model.trained)
        for attribute in ['x_col', 'y_col', 'list_classes', 'dict_classes', 'multi_label', 'level_save', 'batch_size', 'epochs', 'patience',
                          'transformer_name', 'model_max_length']:
            self.assertEqual(getattr(model, attribute), getattr(new_model, attribute))
        for attribute in ['validation_split']:
            self.assertAlmostEqual(getattr(model, attribute), getattr(new_model, attribute))
        self.assertEqual(set(model.trainer_params), set(new_model.trainer_params))
        self.assertEqual(model.trainer_params['evaluation_strategy'], new_model.trainer_params['evaluation_strategy'])
        self.assertAlmostEqual(model.trainer_params['weight_decay'], new_model.trainer_params['weight_decay'])
        self.assertEqual(model.trainer_params['load_best_model_at_end'], new_model.trainer_params['load_best_model_at_end'])
        self.assertEqual(set(model.transformer_params), set(new_model.transformer_params))
        self.assertEqual(model.transformer_params['evaluation_strategy'], new_model.transformer_params['evaluation_strategy'])
        self.assertAlmostEqual(model.transformer_params['weight_decay'], new_model.transformer_params['weight_decay'])
        self.assertEqual(model.transformer_params['load_best_model_at_end'], new_model.transformer_params['load_best_model_at_end'])
        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

    def test15_model_huggingface_load_standalone_files(self):
        '''Test of the method _load_standalone_files of {{package_name}}.models_training.models_tensorflow.model_huggingface.ModelHuggingFace'''
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case with default_model_dir
        model = ModelHuggingFace(model_dir=model_dir, epochs=2)
        x_train = ['coucou', 'coucou_2', 'coucou_3']
        y_train = ['class_1', 'class_2', 'class_1']
        x_valid = ['coucou_4', 'coucou_5', 'coucou_6']
        y_valid = ['class_1', 'class_2', 'class_1']
        model.fit(x_train=x_train, y_train=y_train, x_valid=x_valid, y_valid=y_valid)
        model.save(json_data={'test': 8})

        configs = model.load_configs(model_dir=model_dir)
        new_model = ModelHuggingFace._init_new_instance_from_configs(configs=configs)
        new_model._load_standalone_files(default_model_dir=model_dir)
        self.assertTrue(isinstance(new_model.model, type(model.model)))
        self.assertTrue(isinstance(new_model.tokenizer, type(model.tokenizer)))
        self.assertEqual(model.model.config.n_layers, new_model.model.config.n_layers)
        self.assertEqual(model.model.config.vocab_size, new_model.model.config.vocab_size)
        self.assertEqual(model.model.config.model_type, new_model.model.config.model_type)

        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

        # Nominal case with paths
        model = ModelHuggingFace(model_dir=model_dir, epochs=2)
        x_train = ['coucou', 'coucou_2', 'coucou_3']
        y_train = ['class_1', 'class_2', 'class_1']
        x_valid = ['coucou_4', 'coucou_5', 'coucou_6']
        y_valid = ['class_1', 'class_2', 'class_1']
        model.fit(x_train=x_train, y_train=y_train, x_valid=x_valid, y_valid=y_valid)
        model.save(json_data={'test': 8})

        configs = model.load_configs(model_dir=model_dir)
        new_model = ModelHuggingFace._init_new_instance_from_configs(configs=configs)
        new_model._load_standalone_files(hf_model_dir_path=os.path.join(model_dir, "hf_model"),
                                         hf_tokenizer_dir_path=os.path.join(model_dir, "hf_tokenizer"))
        self.assertTrue(isinstance(new_model.model, type(model.model)))
        self.assertTrue(isinstance(new_model.tokenizer, type(model.tokenizer)))
        self.assertEqual(model.model.config.n_layers, new_model.model.config.n_layers)
        self.assertEqual(model.model.config.vocab_size, new_model.model.config.vocab_size)
        self.assertEqual(model.model.config.model_type, new_model.model.config.model_type)

        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

        # Errors
        model = ModelHuggingFace(model_dir=model_dir, epochs=2)
        x_train = ['coucou', 'coucou_2', 'coucou_3']
        y_train = ['class_1', 'class_2', 'class_1']
        x_valid = ['coucou_4', 'coucou_5', 'coucou_6']
        y_valid = ['class_1', 'class_2', 'class_1']
        model.fit(x_train=x_train, y_train=y_train, x_valid=x_valid, y_valid=y_valid)
        model.save(json_data={'test': 8})

        configs = model.load_configs(model_dir=model_dir)
        new_model = ModelHuggingFace._init_new_instance_from_configs(configs=configs)
        with self.assertRaises(ValueError):
            new_model._load_standalone_files(hf_tokenizer_dir_path=os.path.join(model_dir, "hf_tokenizer"))
        with self.assertRaises(ValueError):
            new_model._load_standalone_files(hf_model_dir_path=os.path.join(model_dir, "hf_model"))
        with self.assertRaises(FileNotFoundError):
            new_model._load_standalone_files(hf_model_dir_path=os.path.join(model_dir, "hf_model2"),
                                         hf_tokenizer_dir_path=os.path.join(model_dir, "hf_tokenizer"))
        with self.assertRaises(FileNotFoundError):
            new_model._load_standalone_files(hf_model_dir_path=os.path.join(model_dir, "hf_model"),
                                         hf_tokenizer_dir_path=os.path.join(model_dir, "hf_tokenizer2"))

        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

        
# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Useful file for the NLP tutorial
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import logging
import numpy as np
import pandas as pd

from {{package_name}} import utils


## Manage logger
# Get logger (def level: INFO)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
# Get console handler
# On log tout ce qui est possible ici (i.e >= level du logger)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
# Manage formatter
formatter = logging.Formatter('[%(asctime)s] - %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
ch.setFormatter(formatter)
# Add handler to the logger
logger.addHandler(ch)


def change_dir():
    '''Gets us to the root of this script'''
    abspath = os.path.abspath(__file__)
    dname = os.path.dirname(abspath)
    os.chdir(dname)


def test_exercice_1():
    '''Tests the first exercise'''
    change_dir()
    data_path = utils.get_data_path()
    train_path = os.path.join(data_path, 'dataset_jvc_train.csv')
    valid_path = os.path.join(data_path, 'dataset_jvc_valid.csv')
    test_path = os.path.join(data_path, 'dataset_jvc_test.csv')
    if not os.path.exists(train_path):
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        logger.error(f"Can't find file {train_path}")
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
    elif not os.path.exists(valid_path):
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        logger.error(f"Can't find file {valid_path}")
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
    elif not os.path.exists(test_path):
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        logger.error(f"Can't find file {test_path}")
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
    # Check the seed
    else:
        df_train = pd.read_csv(train_path, sep=';', encoding='utf-8')
        expected_game = 'Trials Evolution : Gold Edition sur PC'
        game = df_train['game'].iloc[5]
        if expected_game != game:
            print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
            logger.error("The files have been found but with wrong content.")
            logger.error("Did you put the random seed equal to 42?")
            print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        else:
            print(f"\033[1m\033[92m😀 TEST SUCCEEDED 😀\033[0m")
            print(f"\033[1m\033[92m😀 TEST SUCCEEDED 😀\033[0m")
    return None


def get_exercice_1_solution():
    '''Gets the solution for exercise 1'''
    change_dir()
    data_path = utils.get_data_path()
    dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
    dir_path = os.path.split(dir_path)[0]
    scripts_path = os.path.join(dir_path, '{{package_name}}-scripts', 'utils')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")
    print('')
    print('Activate your virtual environment')
    print('')
    print(f"cd {scripts_path}")
    print('')
    print(f"python 0_split_train_valid_test.py -f dataset_jvc.csv --split_type random --seed 42")
    print('')
    print(f"If you already generated some files and need to overwrite them, you need to add the --overwrite argument:")
    print(f"python 0_split_train_valid_test.py -f dataset_jvc.csv --split_type random --seed 42 --overwrite")
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")


def test_exercice_2():
    '''Tests the second exercise'''
    change_dir()
    project_path = os.path.join(os.getcwd(), 'my_awesome_folder')
    data_path = utils.get_data_path()
    train_sample_path = os.path.join(data_path, 'dataset_jvc_train_10_samples.csv')
    if not os.path.exists(train_sample_path):
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        logger.error(f"Can't find file {train_sample_path}")
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
    # Check that there are 10 lines
    else:
        df_train_sample = pd.read_csv(train_sample_path, sep=';', encoding='utf-8')
        expected_n_lines = 10
        n_lines = df_train_sample.shape[0]
        if expected_n_lines != n_lines:
            print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
            logger.error(f"The file was found but with {n_lines} lines instead of {expected_n_lines} !")
            print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        else:
            print(f"\033[1m\033[92m😀 TEST SUCCEEDED 😀\033[0m")
            print(f"\033[1m\033[92m😀 TEST SUCCEEDED 😀\033[0m")
    return None


def get_exercice_2_solution():
    '''Gets the solution for exercise 2'''
    change_dir()
    dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
    dir_path = os.path.split(dir_path)[0]
    scripts_path = os.path.join(dir_path, '{{package_name}}-scripts', 'utils')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")
    print('')
    print('Activate your virtual environment')
    print('')
    print(f"cd {scripts_path}")
    print('')
    print(f"python 0_create_samples.py -f dataset_jvc_train.csv -n 10")
    print('')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")


def test_exercice_3():
    '''Tests exercise 3'''
    change_dir()
    data_path = utils.get_data_path()
    train_preprocess_P1_path = os.path.join(data_path, 'dataset_jvc_train_preprocess_P1.csv')
    if not os.path.exists(train_preprocess_P1_path):
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        logger.error(f"Can't find file {train_preprocess_P1_path}")
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
    # Check one line
    else:
        df_train_preprocess_P1 = pd.read_csv(train_preprocess_P1_path, sep=';', encoding='utf-8', skiprows=1)
        expected_preprocess = 'trials evolution gold edition est un jeu mêlant courses et plates formes sur pc le joueur y pilote une moto dans des niveaux semés d embûches et d obstacles pour les franchir il doit jouer en permanence sur l équilibre de sa monture un éditeur de niveaux permet de créer des pistes et de les partager en ligne cette version pc comprend en outre le jeu trials hd'
        preprocess = df_train_preprocess_P1['preprocessed_text'].iloc[5]
        if expected_preprocess != preprocess:
            print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
            logger.error("The preprocessed file exists but with the wrong data.")
            logger.error("Did you apply preprocess_P1 to dataset_jvc_train.csv ?")
            print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        else:
            print(f"\033[1m\033[92m😀 TEST SUCCEEDED 😀\033[0m")
            print(f"\033[1m\033[92m😀 TEST SUCCEEDED 😀\033[0m")
    return None


def get_exercice_3_solution():
    '''Gets the solution for exercise 3'''
    change_dir()
    dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
    dir_path = os.path.split(dir_path)[0]
    scripts_path = os.path.join(dir_path, '{{package_name}}-scripts')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")
    print('')
    print('Activate your virtual environment')
    print('')
    print(f"cd {scripts_path}")
    print('')
    print(f"python 1_preprocess_data.py -f dataset_jvc_train.csv --input_col description")
    print('')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")


def test_exercice_4():
    '''Tests exercise 4'''
    change_dir()
    data_path = utils.get_data_path()
    train_preprocess_P2_path = os.path.join(data_path, 'dataset_jvc_train_preprocess_P2.csv')
    valid_preprocess_P2_path = os.path.join(data_path, 'dataset_jvc_valid_preprocess_P2.csv')
    if not os.path.exists(train_preprocess_P2_path):
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        logger.error(f"Can't find file {train_preprocess_P2_path}")
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
    elif not os.path.exists(valid_preprocess_P2_path):
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        logger.error(f"Can't find file {valid_preprocess_P2_path}")
        print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
    # Check one line on the train
    else:
        df_train_preprocess_P2 = pd.read_csv(train_preprocess_P2_path, sep=';', encoding='utf-8', skiprows=1)
        expected_preprocess = 'trials evolution gold edition jeu mêlant courses plates formes pc joueur pilote moto niveaux semés embûches obstacles franchir jouer permanence équilibre monture éditeur niveaux créer pistes partager ligne version pc comprend jeu trials hd'
        preprocess = df_train_preprocess_P2['preprocessed_text'].iloc[5]
        if expected_preprocess != preprocess:
            print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
            logger.error("The preprocessed file exists but with the wrong data.")
            logger.error("Did you apply preprocess_P2 to dataset_jvc_train.csv ?")
            print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
        # Check one line on the valid
        else:
            df_valid_preprocess_P2 = pd.read_csv(valid_preprocess_P2_path, sep=';', encoding='utf-8', skiprows=1)
            expected_preprocess = 'jeu plates formes poétique xbox one ori and the blind forest hommage gameplay jeux aventure rpg metroidvania joueur explorer niveaux trouver pouvoirs pouvoir accéder nouvelles zones'
            preprocess = df_valid_preprocess_P2['preprocessed_text'].iloc[17]
            if expected_preprocess != preprocess:
                print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
                logger.error("The preprocessed file exists but with the wrong data.")
                logger.error("Did you apply preprocess_P2 to dataset_jvc_valid.csv ?")
                print(f"\033[1m\033[91m🤬 TEST FAILED 🤬\033[0m")
            else:
                print(f"\033[1m\033[92m😀 TEST SUCCEEDED 😀\033[0m")
                print(f"\033[1m\033[92m😀 TEST SUCCEEDED 😀\033[0m")
    return None


def get_exercice_4_solution():
    '''Gets the solution for exercise 4'''
    change_dir()
    dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
    dir_path = os.path.split(dir_path)[0]
    scripts_path = os.path.join(dir_path, '{{package_name}}-scripts')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")
    print('')
    print('In the file preprocess.py :')
    print('')
    print('   1. Add the function preprocess_sentence_P2 :')
    print('''
@wnf_utils.data_agnostic
@wnf_utils.regroup_data_series
def preprocess_sentence_P2(docs):
    if type(docs) != pd.Series: raise TypeError("L'objet docs doit être du type pd.Series.")
    pipeline = ['remove_non_string', 'get_true_spaces', 'remove_punct', 'to_lower',
                'remove_stopwords', 'trim_string', 'remove_leading_and_ending_spaces']
    return api.preprocess_pipeline(docs, pipeline=pipeline, chunksize=100000)
    ''')
    print('')
    print('   2. Add an entry in the dictionary preprocessors_dict :')
    print('''
            preprocessors_dict = {
                'no_preprocess': lambda x: x,
                'preprocess_P1': preprocess_sentence_P1, # Example of a preprocessing
                'preprocess_P2': preprocess_sentence_P2, # New preprocessing
            }
    ''')
    print('')
    print('Activate your virtual environment')
    print('')
    print(f"cd {scripts_path}")
    print('')
    print(f"python 1_preprocess_data.py -f dataset_jvc_train.csv dataset_jvc_valid.csv --input_col description --preprocessing preprocess_P2")
    print('')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")


def get_exercice_5_solution():
    '''Gets the solution for exercise 5'''
    change_dir()
    dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
    dir_path = os.path.split(dir_path)[0]
    scripts_path = os.path.join(dir_path, '{{package_name}}-scripts')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")
    print('')
    print(f"Activate your virtual environment")
    print('')
    print(f"cd {scripts_path}")
    print('')
    print(f"python 2_training.py --filename dataset_jvc_train_preprocess_P2.csv --x_col preprocessed_text --y_col RPG --filename_valid dataset_jvc_valid_preprocess_P2.csv")
    print('')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")


def get_exercice_6_solution():
    '''Gets the solution for exercise 6'''
    change_dir()
    dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
    dir_path = os.path.split(dir_path)[0]
    scripts_path = os.path.join(dir_path, '{{package_name}}-scripts')
    scripts_path_utils = os.path.join(scripts_path, 'utils')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")
    print('')
    print('# Comment the line with ModelTfidfSvm in 2_training.py.')
    print('')
    print('# Uncomment the line with ModelEmbeddingLstm in 2_training.py.')
    print('')
    print(f"Activate your virtual environment")
    print('')
    print(f"cd {scripts_path_utils}")
    print('')
    print(f"python 0_get_embedding_dict.py -f cc.fr.300.vec")
    print('')
    print(f"cd {scripts_path}")
    print('')
    print(f'python 2_training.py --filename dataset_jvc_train.csv --x_col description --y_col "Action" "Aventure" "RPG" "Plate-Forme" "FPS" "Course" "Strategie" "Sport" "Reflexion" "Combat" --filename_valid dataset_jvc_valid.csv')
    print('')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")


def get_exercice_7_solution():
    '''Gets the solution for exercise 8'''
    change_dir()
    dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
    dir_path = os.path.split(dir_path)[0]
    scripts_path = os.path.join(dir_path, '{{package_name}}-scripts')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")
    print('')
    print(f"cd {scripts_path}")
    print('')
    print(f"Activate your virtual environment")
    print('')
    print('python 3_predict.py --filename dataset_jvc_test.csv --x_col description --y_col "Action" "Aventure" "RPG" "Plate-Forme" "FPS" "Course" "Strategie" "Sport" "Reflexion" "Combat" --model_dir model_embedding_lstm_{YYYY_MM_DD-hh_mm_ss}')
    print('')
    print(f"\033[1m\033[94m💡 SOLUTION 💡\033[0m")



#!/usr/bin/env python3

## Classe modèle pour le tutoriel Jules Verne
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import logging
import os
import re
import math
import numpy as np
import pandas as pd
from typing import List, Union
from collections import Counter

from {{package_name}}.models_training.models_sklearn.model_tfidf_svm import ModelTfidfSvm


class ModelAuthor(ModelTfidfSvm):

    _default_name = 'model_author'

    def __init__(self, nb_word_sentence: int = 50, **kwargs) -> None:
        '''Initialization of the class (see ModelTfidfSvm & ModelPipeline & ModelClass for more arguments)

        Kwargs:
            nb_word_sentence (int) : Number of words for each sentence
        Raises:
            ValueError: If multilabel task
        '''
        super().__init__(**kwargs)
        self.nb_word_sentence = nb_word_sentence  # Number of words per sentence
        # We do not manage multilabel task
        if self.multi_label:
            raise ValueError("On ne gère pas les cas multilabel")

    def fit(self, x_train, y_train, **kwargs) -> None:
        '''Trains the model
           **kwargs permits compatibility with Keras model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
            y_train (?): Array-like, shape = [n_samples, n_targets]
        Raises:
            RuntimeError: If the model is already fitted
        '''
        if self.trained:
            self.logger.error("We can't train again a pipeline sklearn model")
            self.logger.error("Please train a new model")
            raise RuntimeError("We can't train again a pipeline sklearn model")

        # Split texts into sentence / author pairs
        df_train_sentences = df_texts_to_df_sentences(x_train, y_train, nb_word_sentence=self.nb_word_sentence)
        new_x_train = df_train_sentences['sentence']
        new_y_train = df_train_sentences['author']
        # Fit
        super().fit(new_x_train, new_y_train)

    def predict(self, x_test, return_proba: bool = False, **kwargs) -> np.ndarray:
        '''Predictions on test set

        Args:
            x_test (?): Array-like or sparse matrix, shape = [n_samples, n_features]
        Kwargs:
            return_proba (bool): If the function should return the probabilities instead of the classes (Keras compatibility)
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        # Process texts one by one
        list_predictions = []
        for text in x_test:
            # Get predictions for each text' sentences
            sentences = text_to_sentences(text, nb_word_sentence=self.nb_word_sentence)
            predictions = super().predict(sentences)
            # Count nb of occurrences for each author
            counter_author_predictions = dict(Counter(list(predictions)))

            # If we do not return probabilites, just add predicted author (max count)
            if not return_proba:
                predicted_author = max(counter_author_predictions, key=counter_author_predictions.get)
                list_predictions.append(predicted_author)  # On ajoute l'auteur predit à la liste
            else:
                # We consider probabilities as the percentage of sentences predicted
                nb_sentences = len(sentences)
                list_probas = [counter_author_predictions.get(author, 0) / nb_sentences for author in self.list_classes]
                list_predictions.append(list_probas)

        # Return
        return np.array(list_predictions)

    def predict_proba(self, x_test, **kwargs) -> np.ndarray:
        '''Predicts probabilities on the test dataset

        Args:
            x_test (?): Array-like or sparse matrix, shape = [n_samples, n_features]
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        return self.predict(x_test, return_proba=True, **kwargs)

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save model
        if json_data is None:
            json_data = {}
        json_data['nb_word_sentence'] = self.nb_word_sentence

        # Save
        super().save(json_data=json_data)

    @classmethod
    def _init_new_instance_from_configs(cls, configs):
        '''Inits a new instance from a set of configurations

        Args:
            configs: a set of configurations of a model to be reloaded
        Returns:
            ModelClass: the newly generated class
        '''
        # Call parent
        model = super()._init_new_instance_from_configs(configs)

        # Try to read the following attributes from configs and, if absent, keep the current one
        for attribute in ['nb_word_sentence']:
            setattr(model, attribute, configs.get(attribute, getattr(model, attribute)))

        # Return the new model
        return model

### Fonctions utilitaires


def text_to_sentences(text: str, nb_word_sentence: int = 10) -> List[str]:
    '''Transforms a text in sentences.

    Args:
        text (str): The text to cut in sentences
    Kwargs:
        nb_word_sentence (int): The number of words in a sentence
    Returns:
        list: A list of sentences
    '''
    text = re.sub(r'\s', ' ', text)
    # Change some punctuations to a whitespace
    for punctuation in [r'\!', r'\?', r'\;', r'\.', r'\,', r'\:']:
        text = re.sub(punctuation, ' ', text)
    # Get rid of superfluous whitespaces
    text = re.sub(' +', ' ', text)
    words_list = text.split(' ')
    list_sentences = []
    # Cut the text in sentences
    for i in range(math.ceil(len(words_list) / nb_word_sentence)):
        list_sentences.append(' '.join(words_list[i * nb_word_sentence: (i + 1) * nb_word_sentence]))
    return list_sentences


def df_texts_to_df_sentences(texts: Union[list, pd.Series], authors: Union[list, pd.Series], nb_word_sentence: int = 10) -> pd.DataFrame:
    '''Creates a dataframe containaing 'sentences' from a collection of texts with their corresponding authors.

    Args:
        texts (list, pd.Series): a collection of texts
        authors (list, pd.Series): the corresponding authors
    Kwargs:
        nb_word_sentence (int): The number of words in a sentence
    Returns:
        pd.DataFrame: A dataframe containing sentences extracted from the initial texts with their corresponding author
    '''
    list_sentences = []
    for text, author in zip(list(texts), list(authors)):
        # This function transforms a text in 'sentences'
        sentences = text_to_sentences(text, nb_word_sentence)
        list_sentences = list_sentences + [(sentence, author) for sentence in sentences]
    df_sentences = pd.DataFrame(list_sentences, columns=['sentence', 'author'])
    return df_sentences


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import words_n_fun as wnf

# Get logger (def level: INFO)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Get console handler
# We log evetything possible here (ie. >= logger level)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

# Manage formatter
formatter = logging.Formatter("[%(asctime)s] - %(name)s.%(funcName)s() - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
ch.setFormatter(formatter)

# Add handler to the logger
logger.addHandler(ch)


## Manage dependancies loggers (def level: INFO)
wnf.logger.setLevel(logging.INFO)



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest

# Utils libs
import os
import json
import shutil
import mlflow
import pathlib
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from urllib.parse import urlparse

from {{package_name}}.monitoring.mlflow_logger import MLflowLogger

# Disable logging
import logging
logging.disable(logging.CRITICAL)

# TMP directory for mlruns
TMP_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'tmp_experiments')
MLRUNS_DIR = os.path.join(TMP_DIR, 'mlruns')
MLRUNS_ARTIFACT_DIR = os.path.join(TMP_DIR, 'mlruns_artifacts')

LOCAL_TRACKING_URI = pathlib.Path(MLRUNS_DIR).as_uri()
LOCAL_ARTIFACT_URI = pathlib.Path(MLRUNS_ARTIFACT_DIR).as_uri()


class MLflowLoggerTests(unittest.TestCase):
    '''Main class to test mlflow_logger'''

    @classmethod
    def setUpClass(cls):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        # Clean mlruns directory (if exists)
        if os.path.exists(TMP_DIR):
            shutil.rmtree(TMP_DIR)

    @classmethod
    def tearDownClass(cls):
        # Remove mlruns directory
        if os.path.exists(TMP_DIR):
            shutil.rmtree(TMP_DIR)

    def test01_mlflow_logger_init(self):
        '''Test of the initialization of {{package_name}}.monitoring.mlflow_logger.MLflowLogger'''
        experiment_name = 'test_mlflow_logger_init'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        self.assertEqual(mlflow_logger.tracking_uri, LOCAL_TRACKING_URI)
        self.assertEqual(mlflow.get_tracking_uri(), LOCAL_TRACKING_URI)
        self.assertEqual(mlflow_logger.experiment_name, experiment_name)
        # No need to call end_run

    def test02_mlflow_logger_end_run(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.end_run'''
        experiment_name = 'test_mlflow_logger_end_run'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # We activate a run via a log
        mlflow_logger.log_param('stop', 'toto')
        # Use of end_run & check
        mlflow_logger.end_run()
        self.assertEqual(mlflow.active_run(), None)

    def test03_mlflow_logger_log_metric(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.log_metric'''
        experiment_name = 'test_mlflow_logger_log_metric'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        mlflow_logger.log_metric('test', 5)
        mlflow_logger.log_metric('test', 5, step=2)
        
        df_run = mlflow.search_runs(mlflow_logger.experiment_id)
        self.assertTrue("metrics.test" in df_run.columns)
        mlflow_logger.end_run()

    def test04_mlflow_logger_log_metrics(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.log_metrics'''
        experiment_name = 'test_mlflow_logger_log_metrics'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        mlflow_logger.log_metrics({'test': 5, 'test2': 24})
        mlflow_logger.log_metrics({'test': 5}, step=2)

        df_run = mlflow.search_runs(mlflow_logger.experiment_id)
        self.assertTrue("metrics.test" in df_run.columns)
        self.assertTrue("metrics.test2" in df_run.columns)
        mlflow_logger.end_run()

    def test05_mlflow_logger_log_param(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.log_param'''
        experiment_name = 'test_mlflow_logger_log_param'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        mlflow_logger.log_param('test', 5)

        df_run = mlflow.search_runs(mlflow_logger.experiment_id)
        self.assertTrue("params.test" in df_run.columns)
        mlflow_logger.end_run()

    def test06_mlflow_logger_log_params(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.log_params'''
        experiment_name = 'test_mlflow_logger_log_params'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        mlflow_logger.log_params({'test': 5, 'test2': 24})

        df_run = mlflow.search_runs(mlflow_logger.experiment_id)
        self.assertTrue("params.test" in df_run.columns)
        self.assertTrue("params.test2" in df_run.columns)
        mlflow_logger.end_run()

    def test07_mlflow_logger_set_tag(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.set_tag'''
        experiment_name = 'test_mlflow_logger_set_tag'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        mlflow_logger.set_tag('test', 5)

        df_run = mlflow.search_runs(mlflow_logger.experiment_id)
        self.assertTrue("tags.test" in df_run.columns)
        mlflow_logger.end_run()

    def test08_mlflow_logger_set_tags(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.set_tags'''
        experiment_name = 'test_mlflow_logger_set_tags'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        mlflow_logger.set_tags({'test': 5, 'test2': 24})

        df_run = mlflow.search_runs(mlflow_logger.experiment_id)
        self.assertTrue("tags.test" in df_run.columns)
        self.assertTrue("tags.test2" in df_run.columns)
        # Clear
        mlflow_logger.end_run()

    def test09_mlflow_logger_valid_name(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.valid_name'''
        experiment_name = 'test_mlflow_logger_valid_name'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Valid case
        self.assertTrue(mlflow_logger.valid_name('valid_name'))
        self.assertFalse(mlflow_logger.valid_name('not a valid_name!'))
        # No need to call end_run

    def test10_mlflow_logger_log_df_stats(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.log_df_stats'''
        experiment_name = 'test_mlflow_logger_log_df_stats'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)

        # Nominal case
        df_stats = pd.DataFrame({
            'Label': ['label1', 'label2', 'label3!?', 'label4'],
            'metric1': [0.5, 0.2, 1.0, -1.5],
            'metric2!!!': [None, 0.2, -15, np.NaN]
        })
        mlflow_logger.log_df_stats(df_stats)
        
        df_run = mlflow.search_runs(mlflow_logger.experiment_id)

        for param in ('Label 0', 'Label 1', 'Label 2', 'Label 3'):
            self.assertTrue(f"params.{param}" in df_run.columns)
            
        for metric in ('label1 --- metric1', 'label1 --- Col 1', 'label2 --- metric1', 'label2 --- Col 1', 'Label 2 --- metric1', 'Label 2 --- Col 1', 'label4 --- metric1', 'label4 --- Col 1'):
            self.assertTrue(f"metrics.{metric}" in df_run.columns)

        # Bad label_col
        with self.assertRaises(ValueError):
            mlflow_logger.log_df_stats(df_stats, label_col='NotALabelCol')

        # Clear
        mlflow_logger.end_run()

    def test11_mlflow_logger_log_dict(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.log_dict'''
        experiment_name = 'test_mlflow_logger_log_dict'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        dict = {'toto': 'titi', 'tata': 5}
        artifact_file = 'test.json'
        mlflow_logger.log_dict(dict, artifact_file)

        df_run = mlflow.search_runs(mlflow_logger.experiment_id, max_results=1)

        artifact_uri = df_run.loc[0, "artifact_uri"]
        artifact_uri_path = urlparse(artifact_uri).path
        saved_json_path = os.path.join(artifact_uri_path, artifact_file)

        self.assertTrue(os.path.exists(saved_json_path))
        with open(saved_json_path, 'r') as f:
            saved_jason = json.load(f)

        self.assertEqual(saved_jason, dict)
        # Clear
        mlflow_logger.end_run()

    def test12_mlflow_logger_log_text(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.log_text'''
        experiment_name = 'test_mlflow_logger_log_text'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        text = 'This is a test !!!'
        artifact_file = 'test.txt'
        mlflow_logger.log_text(text, artifact_file)

        df_run = mlflow.search_runs(mlflow_logger.experiment_id, max_results=1)

        artifact_uri = df_run.loc[0, "artifact_uri"]
        artifact_uri_path = urlparse(artifact_uri).path
        saved_txt_path = os.path.join(artifact_uri_path, artifact_file)

        self.assertTrue(os.path.exists(saved_txt_path))
        with open(saved_txt_path, 'r') as f:
            saved_txt = f.read()
        self.assertEqual(saved_txt, text)
        # Clear
        mlflow_logger.end_run()

    def test13_mlflow_logger_log_figure(self):
        '''Test of {{package_name}}.monitoring.mlflow_logger.MLflowLogger.log_figure'''
        experiment_name = 'test_mlflow_logger_log_figure'
        mlflow_logger = MLflowLogger(experiment_name=experiment_name, tracking_uri=LOCAL_TRACKING_URI, artifact_uri=LOCAL_ARTIFACT_URI)
        # Nominal case
        plt.pie([0.4, 0.3, 0.3])
        figure = plt.gcf()
        figure.canvas.draw()
        image_from_figure = np.frombuffer(figure.canvas.tostring_rgb(), dtype=np.uint8)
        artifact_file = 'figure.png'
        mlflow_logger.log_figure(figure, artifact_file)

        df_run = mlflow.search_runs(mlflow_logger.experiment_id, max_results=1)

        artifact_uri = df_run.loc[0, "artifact_uri"]
        artifact_uri_path = urlparse(artifact_uri).path
        saved_png_path = os.path.join(artifact_uri_path, artifact_file)

        self.assertTrue(os.path.exists(saved_png_path))
        saved_png = (plt.imread(saved_png_path) * 255).astype('uint8')
        saved_png = saved_png[:, :, :3]
        self.assertAlmostEqual(np.mean(image_from_figure), np.mean(saved_png))
        # # Clear
        mlflow_logger.end_run()

# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3
# type: ignore

# We ignore the type check on mixins : too complicated

## Definition of a parent class for the  object detection models
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelObjectDetectorMixin -> Parent class object_detector


# Cf. fix https://stackoverflow.com/questions/27147300/matplotlib-tcl-asyncdelete-async-handler-deleted-by-the-wrong-thread
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt


import os
import logging
import numpy as np
import pandas as pd
import seaborn as sns
from typing import List, Union
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval

from ... import utils

sns.set(style="darkgrid")


class ModelObjectDetectorMixin:
    '''Parent class (Mixin) for the model of type object detector'''

    # Not implemented :
    # -> predict : to be implemented by the parent class using this mixin

    def __init__(self, level_save: str = 'HIGH', **kwargs) -> None:
        '''Initialization of the parent class - Object detector

        Kwargs:
            level_save (str): Level of saving
                LOW: stats + configurations + logger keras - /!\\ The model can't be reused /!\\ -
                MEDIUM: LOW + hdf5 + pkl + plots
                HIGH: MEDIUM + predictions
        Raises:
            ValueError: If the object level_save is not a valid option (['LOW', 'MEDIUM', 'HIGH'])
        '''
        super().__init__(level_save=level_save, **kwargs)  # forwards level_save & all unused arguments

        if level_save not in ['LOW', 'MEDIUM', 'HIGH']:
            raise ValueError(f"The object level_save ({level_save}) is not a valid option (['LOW', 'MEDIUM', 'HIGH'])")

        # Get logger
        self.logger = logging.getLogger(__name__)

        # Model type
        self.model_type = 'object_detector'

        # List of classes to consider (set on fit)
        self.list_classes = None
        self.dict_classes = None

        # Other options
        self.level_save = level_save

    def inverse_transform(self, y) -> list:
        '''Gets a list of classes from predictions.
        Useless here, used solely for compatibility.

        Args:
            y (?): Array-like, shape = [n_samples, n_features], arrays of 0s and 1s
        Returns:
            (list)
        '''
        return list(y) if isinstance(y, np.ndarray) else y

    def get_and_save_metrics(self, y_true: list, y_pred: list, list_files_x: list = None,
                             type_data: str = '', **kwargs) -> pd.DataFrame:
        '''Gets and saves the metrics of a model

        Args:
            y_true (list): Bboxes list, one entry corresponds to the bboxes of one file - truth
                format bbox : {'class': ..., 'x1': ..., 'y1': ..., 'x2': ..., 'y2': ...}
            y_pred (list): Bboxes list, one entry corresponds to the bboxes of one file - predicted
                format bbox : {'class': ..., 'x1': ..., 'y1': ..., 'x2': ..., 'y2': ..., 'proba': ...}
        Kwargs:
            list_files_x (?): List of input files for the prediction
            type_data (str): Type of the dataset (validation, test, ...)
        Returns:
            pd.DataFrame: The dataframe containing statistics
        '''
        # Manage errors
        if len(y_true) != len(y_pred):
            raise ValueError(f"The size of the two lists (y_true et y_pred) must be equal ({len(y_true)} != {len(y_pred)})")
        if list_files_x is not None and len(y_true) != len(list_files_x):
            raise ValueError(f"The size of the two lists (y_true et list_files_x) must be equal ({len(y_true)} != {len(list_files_x)})")

        # Construction dataframe
        if list_files_x is None:
            df = pd.DataFrame({'y_true': y_true, 'y_pred': y_pred})
        else:
            df = pd.DataFrame({'y_true': y_true, 'y_pred': y_pred, 'file_path': list_files_x})

        # Save a prediction file if wanted
        if self.level_save == 'HIGH':
            file_path = os.path.join(self.model_dir, f"predictions{'_' + type_data if len(type_data) > 0 else ''}.csv")
            if 'file_path' in df.columns:
                df = df.sort_values('file_path')
            df.to_csv(file_path, sep='{{default_sep}}', index=None, encoding='{{default_encoding}}')

        # Print info on missing classes and the impact on metrics
        gt_classes = set([bbox['class'] for bboxes in y_true for bbox in bboxes])
        gt_classes_not_in_model = gt_classes.difference(set(self.list_classes))
        model_classes_not_in_gt = set(self.list_classes).difference(gt_classes)
        # Prints
        if len(gt_classes_not_in_model):
            self.logger.info(f"Classes {gt_classes_not_in_model} are not predicted by the model.")
            self.logger.info("We won't take them into account in the calculation of the metrics.")
        if len(model_classes_not_in_gt):
            self.logger.info(f"Classes {model_classes_not_in_gt} are not present in the dataset used to calculate the metrics.")
            self.logger.info("Metrics on these classes won't be accurate.")

        # Get the classes support
        total_bbox = sum([1 for image in y_true for bbox in image if bbox['class'] in self.list_classes])
        classes_support = {}
        if total_bbox == 0:
            total_bbox = 1
        for cl in self.list_classes:
            classes_support[cl] = sum([bbox['class'] == cl for image in y_true for bbox in image]) / total_bbox

        # Get metrics
        # We use the COCO method to get the Average Precision (AP)
        dict_ap_coco = self._get_coco_ap(y_true, y_pred)

        # Calculate the mean Average Precision (mAP) (weighted or not)
        coco_map = np.mean([value for value in list(dict_ap_coco.values()) if not np.isnan(value)])
        coco_wap = sum([dict_ap_coco[cl] * classes_support[cl] for cl in self.list_classes if classes_support[cl] > 0])

        # Global statistics
        self.logger.info('-- * * * * * * * * * * * * * * --')
        self.logger.info(f"Statistics mAP{' ' + type_data if len(type_data) > 0 else ''}")
        self.logger.info('--------------------------------')
        self.logger.info(f"mean Average Precision (mAP) - COCO method : {round(coco_map, 4)}")
        self.logger.info('--------------------------------')
        self.logger.info(f"weighted Average Precision (wAP) - COCO method : {round(coco_wap, 4)}")
        self.logger.info('--------------------------------')

        # Statistics per classes
        for cl in self.list_classes:
            self.logger.info(f"Class {cl}: AP COCO = {round(dict_ap_coco[cl], 4)} /// Support = {round(classes_support[cl], 4)}")
        self.logger.info('--------------------------------')

        # Construction df_stats
        df_stats = pd.DataFrame(columns=['Label', 'AP COCO', 'Support'])
        df_stats = df_stats.append({'Label': 'All', 'AP COCO': coco_map, 'Support': 1.0}, ignore_index=True)
        for cl in self.list_classes:
            df_stats = df_stats.append({'Label': cl, 'AP COCO': dict_ap_coco[cl], 'Support': classes_support[cl]}, ignore_index=True)

        # Save csv
        file_path = os.path.join(self.model_dir, f"map_coco{'_' + type_data if len(type_data) > 0 else ''}@{round(coco_map, 4)}.csv")
        df_stats.to_csv(file_path, sep='{{default_sep}}', index=False, encoding='{{default_encoding}}')

        # Return df_stats
        return df_stats

    def _get_coco_ap(self, y_true: list, y_pred: list) -> dict:
        '''Calculate COCO's AP for each of the class and gives the result in a dictionary
        where the keys are the classes and the valeus, the corresponding AP value

         Args:
            y_true (list): Bboxes list, one entry corresponds to the bboxes of one file - truth
                format bbox : {'class': ..., 'x1': ..., 'y1': ..., 'x2': ..., 'y2': ...}
            y_pred (list): Bboxes list, one entry corresponds to the bboxes of one file - predicted
                format bbox : {'class': ..., 'x1': ..., 'y1': ..., 'x2': ..., 'y2': ..., 'proba': ...}
        Returns:
            The dictionary containing AP for each class
        '''
        inv_dict_classes = {value: key for key, value in self.dict_classes.items()}
        # Put the bboxes in COCO format
        coco_true = self._put_bboxes_in_coco_format(y_true, inv_dict_classes)
        coco_pred = self._put_bboxes_in_coco_format(y_pred, inv_dict_classes)
        images = [{'id': i + 1} for i in range(len(y_true))]
        categories = [{'id': class_id, 'name': class_name, 'supercategory': 'none'}
                      for class_id, class_name in self.dict_classes.items()]
        dataset_coco_true = {'type': 'instances',
                             'images': images.copy(),
                             'categories': categories.copy(),
                             'annotations': coco_true}
        dataset_coco_pred = {'images': images.copy(),
                             'categories': categories.copy(),
                             'annotations': coco_pred}
        # Call pycocotools API to calculate the AP
        coco_eval = self._get_coco_evaluations(dataset_coco_true, dataset_coco_pred)
        dict_ap = self._get_ap_for_classes(coco_eval)
        return dict_ap

    @classmethod
    def _put_bboxes_in_coco_format(self, bboxes: List[List[dict]], inv_dict_classes: dict) -> List[dict]:
        '''Puts a list of list of bboxes (for example from a prediction) in the right format for pycocotools API.

        Args:
            bboxes (list<list<dict>>) : A list of list of bboxes. The first level of list corresponds to the images and the second level to the
            bboxes of this image.
            inv_dict_classes (dict) : The dictionary of classes in the format {class_name: class_id}
        Returns:
            A list of bboxes
        '''
        annotations = []
        idx_bbox = 1  # WARNING: index begins at 1
        for idx_img, list_bboxes in enumerate(bboxes):
            for bbox in list_bboxes:
                dict_bbox = {'id': idx_bbox,
                             'image_id': idx_img + 1,  # WARNING : index begins at 1
                             'category_id': inv_dict_classes[bbox['class']],
                             'bbox': np.array([bbox['x1'], bbox['y1'], bbox['x2'] - bbox['x1'], bbox['y2'] - bbox['y1']]),
                             'area': (bbox['y2'] - bbox['y1']) * (bbox['x2'] - bbox['x1']),
                             'iscrowd': 0,
                             'score': bbox.get('proba', 1)}
                idx_bbox += 1
                annotations.append(dict_bbox.copy())
        return annotations

    @classmethod
    def _get_coco_evaluations(self, dataset_coco_true: dict, dataset_coco_pred: dict) -> COCOeval:
        '''Calculates the AP from true and predicted datasets in the COCO format, the returns COCOeval,
        the pycocotools API containing all the results.

        Args:
            dataset_coco_true (dict) : Ground truth bboxes in COCO format
            dataset_coco_pred (dict) : Predicted bboxes in COCO format
        Returns:
            A COCOeval (pycocotools API) containing the AP
        '''
        # Everything on mute ! pycocotools library prints too much logs and there are no level settings
        with utils.HiddenPrints():
            # Put the ground truth bboxes in the pycocotools API
            coco_ds = COCO()
            coco_ds.dataset = dataset_coco_true.copy()
            coco_ds.createIndex()

            # Put the predicted bboxes in the pycocotools API
            coco_dt = COCO()
            coco_dt.dataset = dataset_coco_pred.copy()
            coco_dt.createIndex()

            # Get image IDs
            imgIds = sorted(coco_ds.getImgIds())

            # Set evaluator
            cocoEval = COCOeval(coco_ds, coco_dt, 'bbox')
            cocoEval.params.imgIds = imgIds
            cocoEval.params.useCats = True
            cocoEval.params.iouType = "bbox"

            # Evaluate
            cocoEval.evaluate()
            cocoEval.accumulate()

        # Return evaluator
        return cocoEval

    def _get_ap_for_classes(self, coco_eval: COCOeval) -> dict:
        '''Gets the AP per class from cocoEval, the pycocotools API.

        Args:
            coco_eval (COCOeval) : A pycocotools COCOeval which calculated the AP.
                In this function, we just get them, we do not calculate them
        Raises:
            ValueError : The precision has not the right shape (iou, recall, cls, area range, max dets)
        Returns:
            The dictionary containing the AP for each class
        '''
        # Compute per-category AP
        # from https://detectron2.readthedocs.io/en/latest/_modules/detectron2/evaluation/coco_evaluation.html
        precisions = coco_eval.eval["precision"]
        # precision has dims (iou, recall, cls, area range, max dets)
        if len(self.dict_classes) != precisions.shape[2]:
            raise ValueError(f"The precision has not the right shape (iou, recall, cls, area range, max dets): {precisions.shape}")

        # Retrieve APs
        dict_ap = {}
        for idx, name in self.dict_classes.items():
            # area range index 0: all area ranges
            # max dets index -1: typically 100 per image
            precision = precisions[:, :, idx, 0, -1]
            precision = precision[precision > -1]
            ap = np.mean(precision) if precision.size else float("nan")
            dict_ap[name] = ap
        return dict_ap

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save model
        if json_data is None:
            json_data = {}

        json_data['list_classes'] = self.list_classes
        json_data['dict_classes'] = self.dict_classes

        # Save
        super().save(json_data=json_data)


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Classes to explain models predictions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - Explainer -> Parent class for the explainers
# - ShapExplainer -> Shap Explainer wrapper class

import shap
import base64
import logging
import tempfile
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Type, Union, Any

from ..models_training import utils_models
from ..models_training.model_class import ModelClass


class Explainer:
    '''Parent class for the explainers'''

    def __init__(self, *args, **kwargs) -> None:
        '''Initialization of the parent class'''
        self.logger = logging.getLogger(__name__)

    def explain_instance(self, content: pd.DataFrame, **kwargs) -> Any:
        '''Explains a prediction

        Args:
            content (pd.DataFrame): Single entry to be explained
        Returns:
            (?): An explanation object
        '''
        raise NotImplementedError("'explain_instance' needs to be overridden")

    def explain_instance_as_html(self, content: pd.DataFrame, **kwargs) -> str:
        '''Explains a prediction - returns an HTML object

        Args:
            content (pd.DataFrame): Single entry to be explained
        Returns:
            str: An HTML code with the explanation
        '''
        raise NotImplementedError("'explain_instance_as_html' needs to be overridden")

    def explain_instance_as_json(self, content: pd.DataFrame, **kwargs) -> Union[dict, list]:
        '''Explains a prediction - returns an JSON serializable object

        Args:
            content (str): Text to be explained
        Returns:
            Union[dict, list]: A JSON serializable object containing the explanation
        '''
        raise NotImplementedError("'explain_instance_as_json' needs to be overridden")

class ShapExplainer(Explainer):
    '''Shap Explainer wrapper class'''

    def __init__(self, model: Type[ModelClass], anchor_data: pd.DataFrame, anchor_preprocessed: bool = False) -> None:
        ''' Initialization

        Args:
            model: A model instance with predict (regressors) or predict_proba (classifiers) functions
            anchor_data (pd.DataFrame): data anchor needed by shap (usually 100 data points)
        Kwargs:
            anchor_preprocessed (bool): If the anchor data has already been preprocessed
        Raises:
            TypeError: If the provided model is a regressor and does not implement a `predict` function
            TypeError: If the provided model is a classifier and does not implement a `predict_proba` function
        '''
        super().__init__()
        pred_op = getattr(model, "predict", None)
        pred_proba_op = getattr(model, "predict_proba", None)

        if model.model_type == 'regressor':
            if pred_op is None or not callable(pred_op):
                raise TypeError("The supplied model must implement a predict() function")
        # Check classifier
        if model.model_type == 'classifier':
            if pred_proba_op is None or not callable(pred_proba_op):
                raise TypeError("The supplied model must implement a predict_proba() function")

        # Set attributes
        self.model = model
        self.model_type = model.model_type
        # Our explainers will explain a prediction for a given class / label
        # These atributes are set on the fly and will change the proba function used by the explainer
        self.current_class_or_label_index = 0
        fn_output = self.classifier_fn if self.model_type == 'classifier' else self.regressor_fn

        # Preprocess the anchor data
        if not anchor_preprocessed:
            if self.model.preprocess_pipeline is not None:
                anchor_prep = utils_models.apply_pipeline(anchor_data, self.model.preprocess_pipeline)
            else:
                anchor_prep = anchor_data.copy()
        else:
            # Check columns
            try:
                anchor_prep = anchor_data[self.model.x_col]
            except:
                raise ValueError("Provided anchor data (already preprocessed) do not match model's inputs columns")
        # Create the explainer
        self.explainer = shap.Explainer(fn_output, anchor_prep)

    def classifier_fn(self, content_prep: pd.DataFrame) -> np.ndarray:
        '''Function to get probabilities from a dataset (already preprocessed) - classifiers

        Args:
            content_prep (pd.DataFrame): dataset (already preprocessed) to be considered
        Returns:
            np.array: probabilities
        '''
        # Get probabilities
        # Mypy raises a false error here, needs to be ignored
        return self.model.predict_proba(content_prep)[:, self.current_class_or_label_index]  # type: ignore

    def regressor_fn(self, content_prep: pd.DataFrame) -> np.ndarray:
        '''Function to get predictions from a dataset (already preprocessed) - regressors

        Args:
            content_prep (pd.DataFrame): dataset (already preprocessed) to be considered
        Returns:
            np.array: predictions
        '''
        # Get predictions
        # Mypy raises a false error here, needs to be ignored
        return self.model.predict(content_prep)  # type: ignore

    def explain_instance(self, content: pd.DataFrame, class_or_label_index: Union[int, None] = None, **kwargs) -> shap.Explanation:
        '''Explains predictions by returning a shap.Explanation object

        This function calls the Shap module.

        Args:
            content (pd.DataFrame): Entries to be explained
        Kwargs:
            class_or_label_index (int): for classification only. Class or label index to be considered.
        Returns:
            shap.Explanation: Shap Explanation object
        '''
        # Apply preprocessing
        if self.model.preprocess_pipeline is not None:
            df_prep = utils_models.apply_pipeline(content, self.model.preprocess_pipeline)
        else:
            df_prep = content.copy()
            logger.warning("No preprocessing pipeline found - we consider no preprocessing, but it should not be so !")
        # Set index (if needed)
        if class_or_label_index is not None:
            self.current_class_or_label_index = class_or_label_index
        # Get explanations
        return self.explainer(df_prep)  # Shap values

    def explain_instance_as_html(self, content: pd.DataFrame, class_or_label_index: Union[int, None] = None, **kwargs) -> str:
        '''Explains a prediction - returns an HTML object

        Args:
            content (pd.DataFrame): Single entry to be explained
        Kwargs:
            class_or_label_index (int): for classification only. Class or label index to be considered.
        Returns:
            str: An HTML code with the explanation
        '''
        shap_values = self.explain_instance(content, class_or_label_index=class_or_label_index)
        # Waterfall figure
        plt.clf()
        waterfall_fig = shap.plots.waterfall(shap_values[0], show=False)
        with tempfile.TemporaryFile('w+b') as plt_file:
            waterfall_fig.savefig(plt_file, format='png', bbox_inches='tight')
            plt_file.seek(0)
            encoded = base64.b64encode(plt_file.read())
        plt.clf()
        html_waterfall = f"<img src=\'data:image/png;base64, {encoded.decode('utf-8')}\' class=\"shap_fig\" >"  # Class name is used in the demonstrator
        # Force figure
        html_force = shap.plots.force(shap_values[0]).html()
        # Combine & return
        final_html = f"<head>{shap.getjs()}</head>{html_waterfall}<br><body>{html_force}</body>"
        return final_html

    def explain_instance_as_json(self, content: pd.DataFrame, class_or_label_index: Union[int, None] = None, **kwargs) -> Union[dict, list]:
        '''Explains predictions by returning a JSON serializable object

        This function calls the Shap module.

        Args:
            content (pd.DataFrame): entries to be explained
        Kwargs:
            class_or_label_index (int): for classification only. Class or label index to be considered.
        Returns:
            (Union[dict, list]): Shap values
        '''
        return [
            {
                "features": explanation.feature_names, 
                "preprocessed_values": explanation.data, 
                "shap_values": explanation.values, 
                "shap_base_values": explanation.base_values,
            }
            for explanation in self.explain_instance(content, class_or_label_index=class_or_label_index, **kwargs)
        ]


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



import html
import json
import os
from pathlib import Path

import pandas as pd
from IPython import display
from PIL import Image
from sklearn.compose import ColumnTransformer
from {{package_name}}.utils import get_data_path, get_models_path

DATA_PATH = Path(get_data_path())
MODELS_PATH = Path(get_models_path())

DATASET_NAME = "wine"
DATASET_REG_NAME = "wine_reg"


def verify_exercice_1():
    """Verify first exercice"""
    files = {
        kind: DATA_PATH / f"{DATASET_NAME}_{kind}.csv"
        for kind in ("train", "valid", "test")
    }
    expected_results = {
        "train": {"alcohol": 13.64, "length": 107},
        "valid": {"alcohol": 13.72, "length": 36},
        "test": {"alcohol": 14.37, "length": 35},
    }

    # Verify that files exist
    for file_path in files.values():
        if not file_path.exists():
            raise FileNotFoundError(f"{file_path} not found. Did you use {{package_name}}-scripts/utils/0_split_train_valid_test.py ?")

    # Verify files content
    for file, file_path in files.items():
        df = pd.read_csv(file_path, sep=";")
        if df.shape[0] != expected_results[file]["length"]:
            raise ValueError(
                f"Unexpected length for {file_path}.  "
                f"Did you correctly use a 0.6 / 0.2 / 0.2 split ?"
            )
        if df["alcohol"].iloc[0] != expected_results[file]["alcohol"]:
            raise ValueError(
                f"Unexpected value in {file_path}.  "
                f"Did you correctly use 42 as seed ?"
            )

    print("Exercice 1 : OK ✔")


def verify_exercice_2():
    """Verify second exercice"""
    if os.path.exists(DATA_PATH / f"{DATASET_NAME}_10_samples.csv"):
        print("Exercice 2 : OK ✔")
    else:
        print("Exercice 2 : NOT OK")


def verify_exercice_3():
    """Verify third exercice"""
    report_path = DATA_PATH / "reports" / "sweetviz"
    reports = list(os.walk(report_path))[0][2]
    assert len(reports), f"No report found in {report_path}"

    full_report_path = report_path / reports[-1]

    with full_report_path.open("r") as f:
        report_content = f.read()

    report_html = html.escape(report_content)
    iframe = (
        f'<iframe width="100%" height="750px" srcdoc="{report_html}" '
        f'frameborder="0" allowfullscreen></iframe>'
    )

    display.display_html(iframe, raw=True)

    print("Exercice 3 : OK ✔")


def verify_exercice_4():
    """Verify fourth exercice"""
    file_path = DATA_PATH / f"{DATASET_NAME}_preprocess_P1.csv"
    if not os.path.exists(file_path):
        raise FileNotFoundError("Did you run 1_preprocess_data.py ?")

    df = pd.read_csv(file_path, sep=";", skiprows=1)
    if not df["target"].apply(lambda x: x % 1 == 0).all():
        raise ValueError(
            "target should only contain int. Did you use '--target_cols target' ?"
        )

    print("Exercice 4 : OK ✔")


def verify_exercice_5():
    """Verify fifth exercice"""
    try:
        from {{package_name}}.preprocessing.preprocess import preprocess_P2
    except ImportError:
        raise ImportError(
            "Did you create preprocess_P2 in {{package_name}}.preprocessing.preprocess ?"
        )

    from {{package_name}}.preprocessing.preprocess import get_pipelines_dict

    try:
        if not isinstance(get_pipelines_dict()["preprocess_P2"], ColumnTransformer):
            raise ValueError("Value associated to 'preprocess_P2' in get_pipelines_dict should be 'preprocess_P2()'")
    except KeyError:
        raise KeyError("Did you add 'preprocess_P2' to get_pipelines_dict ?")
    
    file_path = DATA_PATH / f"{DATASET_NAME}_preprocess_P2.csv"
    if not file_path.exists():
        raise FileNotFoundError("Did you run 1_preprocess_data.py with argument '-p preprocess_P2' ?")


    df = pd.read_csv(file_path, sep=";", skiprows=1)
    if not df["target"].apply(lambda x: x % 1 == 0).all():
        raise ValueError(
            "target should only contain int. Did you use '--target_cols target' ?"
        )

    print("Exercice 5 : OK ✔")


def verify_exercice_6():
    """Verify sixth exercice"""
    train_P1 = DATA_PATH / f"{DATASET_NAME}_train_preprocess_P1.csv"
    valid_P1 = DATA_PATH / f"{DATASET_NAME}_valid_preprocess_P1.csv"

    for file, error_msg in (
        (train_P1, "Did you run 1_preprocess_data.py on wine_train.csv ? "),
        (valid_P1, "Did you run 2_apply_existing_pipeline.py on wine_valid.csv ? "),
    ):
        if not file.exists():
            raise FileNotFoundError(error_msg)        

    expected_results = {
        train_P1: 0.858,
        valid_P1: 0.964,
    }

    for file, alcohol_first_value in expected_results.items():
        df = pd.read_csv(file, sep=";", skiprows=1)
        
        if not abs(df["num__alcohol"].iloc[0] - alcohol_first_value) < 1e-3:
            raise ValueError(
                f"Unexpected value in {file}. Did you use 1_preprocess_data.py on "
                f"{DATASET_NAME}_train.csv and 2_apply_existing_pipeline.py on "
                f"{DATASET_NAME}_valid.csv and {DATASET_NAME}_test.csv ?"
            )
            
        if not df["target"].apply(lambda x: x % 1 == 0).all():
            raise ValueError(
                f"Target column in {file} should only contain int. "
                f"Did you use '--target_cols target' ?"
            )

    print("Exercice 6 : OK ✔")


def verify_exercice_7():
    """Verify seventh exercice"""

    models_folders = sorted(
        MODELS_PATH.glob("model_ridge_classifier/model_ridge_classifier_*")
    )

    if not models_folders:
        raise AssertionError(
            "No model_ridge_classifier found. Did you run 3_training_classification.py ?"
        )

    last_model = models_folders[-1]
    last_model_config = last_model / "configurations.json"

    with last_model_config.open("r") as f:
        config = json.load(f)

    filename_valid = config.get("filename_valid", None)
    if filename_valid != f"{DATASET_NAME}_valid_preprocess_P1.csv":
        raise AssertionError(
            f"'filename_valid' is {filename_valid} but should be "
            f"'{DATASET_NAME}_valid_preprocess_P1.csv'. "
            f"Did you use '--filename_valid {DATASET_NAME}_valid_preprocess_P1.csv' ?"
        )

    print("Exercice 7 : OK ✔")

    confusion_matrix = last_model / "plots" / "valid_confusion_matrix.png"

    print("A confusion matrix plot has been automatically produced : \n")

    confusion_matrixt_img = Image.open(confusion_matrix)
    confusion_matrixt_img.show()

    print("\n", confusion_matrix)


def verify_exercice_8():
    """Verify eighth exercice"""

    models_folders = sorted(
        MODELS_PATH.glob("model_lgbm_classifier/model_lgbm_classifier_*")
    )

    if not models_folders:
        raise AssertionError(
            "No model_lgbm_classifier found. Did you properly uncomment ModelLGBMClassifier model ?"
        )

    last_model = models_folders[-1]
    last_model_config = last_model / "configurations.json"

    with last_model_config.open("r") as f:
        config = json.load(f)

    filename_valid = config.get("filename_valid", None)
    if filename_valid != f"{DATASET_NAME}_valid_preprocess_P1.csv":
        raise AssertionError(
            f"'filename_valid' is {filename_valid} but should be "
            f"'{DATASET_NAME}_valid_preprocess_P1.csv'. "
            f"Did you use '--filename_valid {DATASET_NAME}_valid_preprocess_P1.csv' ?"
        )

    print("Exercice 8 : OK ✔")

    confusion_matrix = last_model / "plots" / "valid_confusion_matrix.png"

    print("A confusion matrix plot has been automatically produced : \n")

    confusion_matrixt_img = Image.open(confusion_matrix)
    confusion_matrixt_img.show()

    print("\n", confusion_matrix)


def verify_exercice_9():
    """Verify nineth exercice"""
    predictions_folder = DATA_PATH / "predictions" / f"{DATASET_NAME}_test"

    if not predictions_folder.exists():
        raise AssertionError(
            f"No folder {predictions_folder}. "
            f"Did you run 4_predict.py on {DATASET_NAME}_test ?"
        )

    predictions = sorted(predictions_folder.glob("predictions_*"))

    if not predictions:
        raise AssertionError(
            f"No prediction found. Did you run 4_predict.py on {DATASET_NAME}_test ?"
        )

    predictions_found = False
    for last_predictions in predictions:
        last_predictions_config = last_predictions / "configurations.json"

        with last_predictions_config.open("r") as f:
            config = json.load(f)

        if config["model_name"] == "model_ridge_classifier":
            predictions_found = True
            break

    if not predictions_found:
        raise AssertionError(
            "No predictions found for model_ridge_classifier. "
            "Did you use this model to make your predictions ?"
        )

    print("Exercice 9 : OK ✔")

    confusion_matrix = last_predictions / "plots" / "with_y_true_confusion_matrix.png"

    print("A confusion matrix plot has been automatically produced : \n")

    confusion_matrixt_img = Image.open(confusion_matrix)
    confusion_matrixt_img.show()

    print("\n", confusion_matrix)


def verify_exercice_10():
    """Verify tenth exercice"""
    # Verify presence of train, valid test splits
    for split in ("train", "valid", "test"):
        split_path = DATA_PATH / f"{DATASET_REG_NAME}_{split}.csv"
        if not split_path.exists():
            raise FileNotFoundError(
                f"{split_path} not found. Did you run 0_split_train_valid_test.py ?"
            )

    # Verify preprocessing of train data
    train_preprocess_path = DATA_PATH / f"{DATASET_REG_NAME}_train_preprocess_P3.csv"
    if not train_preprocess_path.exists():
        raise FileNotFoundError(
            f"{train_preprocess_path} not found. Did you run preprocess_P3 pipeline "
            f"on {DATASET_REG_NAME}_train.csv thanks to 1_preprocess_data.py ?"
        )

    # Verify preprocess of validation data
    valid_preprocess_path = DATA_PATH / f"{DATASET_REG_NAME}_valid_preprocess_P3.csv"
    if not valid_preprocess_path.exists():
        raise FileNotFoundError(
            f"{valid_preprocess_path} not found. Did you run previous preprocess_P3 "
            f"pipeline on {DATASET_REG_NAME}_valid.csv thanks to "
            f"2_apply_existing_pipeline.py ?"
        )

    # Verify training
    models_folders = sorted(
        MODELS_PATH.glob("model_knn_regressor/model_knn_regressor_*")
    )

    if not models_folders:
        raise AssertionError(
            "No model_knn_regressor found. Did you properly uncomment "
            "ModelKNNRegressor model and run 3_training_regression.py ?"
        )

    last_model = models_folders[-1]
    last_model_config = last_model / "configurations.json"

    with last_model_config.open("r") as f:
        config = json.load(f)

    filename_valid = config.get("filename_valid", None)
    if filename_valid != f"{DATASET_REG_NAME}_valid_preprocess_P3.csv":
        raise AssertionError(
            f"'filename_valid' is {filename_valid} but should be "
            f"'{DATASET_REG_NAME}_valid_preprocess_P3.csv'. "
            f"Did you use '--filename_valid {DATASET_REG_NAME}_valid_preprocess_P3.csv' ?"
        )

    # Verify predictions
    predictions_folder = DATA_PATH / "predictions" / f"{DATASET_REG_NAME}_test"

    if not predictions_folder.exists():
        raise AssertionError(
            f"No folder {predictions_folder}. "
            f"Did you run 4_predict.py on {DATASET_REG_NAME}_test ?"
        )

    predictions = sorted(predictions_folder.glob("predictions_*"))

    if not predictions:
        raise AssertionError(
            f"No prediction found. Did you run 4_predict.py on {DATASET_REG_NAME}_test ?"
        )

    predictions_found = False
    for last_predictions in predictions:
        last_predictions_config = last_predictions / "configurations.json"

        with last_predictions_config.open("r") as f:
            config = json.load(f)

        if config["model_name"] == "model_knn_regressor":
            predictions_found = True
            break

    if not predictions_found:
        raise AssertionError(
            "No predictions found for model_knn_regressor. "
            "Did you use this model to make your predictions ?"
        )

    print("Exercice 10 : OK ✔")

    error_plot = last_predictions / "plots" / "with_y_true_errors.png"

    print(
        "A plot of predicted values against actual values has been automatically produced :"
    )

    error_plot_img = Image.open(error_plot)
    error_plot_img.show()

    print(error_plot)



#!/usr/bin/env python3

## Dense model - Classification
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelDenseClassifier -> Dense model for classifiction


import os
import json
import shutil
import logging
import dill as pickle
from typing import Union, List, Callable

from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import load_model as load_model_keras
from tensorflow.keras.layers import ELU, BatchNormalization, Dense, Dropout, Input

from ... import utils_deep_keras
from ...model_keras import ModelKeras
from ..model_classifier import ModelClassifierMixin  # type: ignore


class ModelDenseClassifier(ModelClassifierMixin, ModelKeras):
    '''Dense model for classification'''

    _default_name = 'model_dense_classifier'

    def __init__(self, **kwargs) -> None:
        '''Initialization of the class (see ModelClass, ModelKeras & ModelClassifierMixin for more arguments)'''
        # Init.
        super().__init__(**kwargs)

        # Get logger (must be done after super init)
        self.logger = logging.getLogger(__name__)

    def _get_model(self) -> Model:
        '''Gets a model structure - returns the instance model instead if already defined

        Returns:
            (Model): a Keras model
        '''
        # Return model if already set
        if self.model is not None:
            return self.model

        # Get input/output dimensions
        input_dim = len(self.x_col)
        num_classes = len(self.list_classes)

        # Process
        input_layer = Input(shape=(input_dim,))

        x = Dense(64, activation=None, kernel_initializer="he_uniform")(input_layer)
        x = BatchNormalization(momentum=0.9)(x)
        x = ELU(alpha=1.0)(x)
        x = Dropout(0.2)(x)

        x = Dense(64, activation=None, kernel_initializer="he_uniform")(x)
        x = BatchNormalization(momentum=0.9)(x)
        x = ELU(alpha=1.0)(x)
        x = Dropout(0.2)(x)

        # Last layer
        activation = 'sigmoid' if self.multi_label else 'softmax'
        out = Dense(num_classes, activation=activation, kernel_initializer='glorot_uniform')(x)

        # Set model
        model = Model(inputs=input_layer, outputs=[out])

        # Set optimizer
        lr = self.keras_params['learning_rate'] if 'learning_rate' in self.keras_params.keys() else 0.001
        decay = self.keras_params['decay'] if 'decay' in self.keras_params.keys() else 0.0
        self.logger.info(f"Learning rate: {lr}")
        self.logger.info(f"Decay: {decay}")
        optimizer = Adam(lr=lr, decay=decay)

        # Set loss & metrics
        loss = utils_deep_keras.f1_loss if self.multi_label else 'categorical_crossentropy'
        metrics: List[Union[str, Callable]] = ['accuracy'] if not self.multi_label else ['categorical_accuracy', 'categorical_crossentropy', utils_deep_keras.f1, utils_deep_keras.precision, utils_deep_keras.recall, utils_deep_keras.f1_loss]  # type: ignore

        # Compile model
        model.compile(optimizer=optimizer, loss=loss, metrics=metrics)
        if self.logger.getEffectiveLevel() < logging.ERROR:
            model.summary()

        # Try to save model as png if level_save > 'LOW'
        if self.level_save in ['MEDIUM', 'HIGH']:
            self._save_model_png(model)

        # Return
        return model

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save configuration JSON
        if json_data is None:
            json_data = {}

        # Save
        super().save(json_data=json_data)

    def reload_from_standalone(self, **kwargs) -> None:
        '''Reloads a model from its configuration and "standalones" files
        - /!\\ Experimental /!\\ -

        Kwargs:
            configuration_path (str): Path to configuration file
            hdf5_path (str): Path to hdf5 file
            preprocess_pipeline_path (str): Path to preprocess pipeline
        Raises:
            ValueError: If configuration_path is None
            ValueError: If hdf5_path is None
            ValueError: If preprocess_pipeline_path is None
            FileNotFoundError: If the object configuration_path is not an existing file
            FileNotFoundError: If the object hdf5_path is not an existing file
            FileNotFoundError: If the object preprocess_pipeline_path is not an existing file
        '''
        # Retrieve args
        configuration_path = kwargs.get('configuration_path', None)
        hdf5_path = kwargs.get('hdf5_path', None)
        preprocess_pipeline_path = kwargs.get('preprocess_pipeline_path', None)

        # Checks
        if configuration_path is None:
            raise ValueError("The argument configuration_path can't be None")
        if hdf5_path is None:
            raise ValueError("The argument hdf5_path can't be None")
        if preprocess_pipeline_path is None:
            raise ValueError("The argument preprocess_pipeline_path can't be None")
        if not os.path.exists(configuration_path):
            raise FileNotFoundError(f"The file {configuration_path} does not exist")
        if not os.path.exists(hdf5_path):
            raise FileNotFoundError(f"The file {hdf5_path} does not exist")
        if not os.path.exists(preprocess_pipeline_path):
            raise FileNotFoundError(f"The file {preprocess_pipeline_path} does not exist")

        # Load confs
        with open(configuration_path, 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        # Can't set int as keys in json, so need to cast it after reloading
        # dict_classes keys are always ints
        if 'dict_classes' in configs.keys():
            configs['dict_classes'] = {int(k): v for k, v in configs['dict_classes'].items()}
        elif 'list_classes' in configs.keys():
            configs['dict_classes'] = {i: col for i, col in enumerate(configs['list_classes'])}

        # Set class vars
        # self.model_name = # Keep the created name
        # self.model_dir = # Keep the created folder
        self.nb_fit = configs.get('nb_fit', 1)  # Consider one unique fit by default
        self.trained = configs.get('trained', True)  # Consider trained by default
        # Try to read the following attributes from configs and, if absent, keep the current one
        for attribute in ['model_type', 'x_col', 'y_col', 'columns_in', 'mandatory_columns',
                          'list_classes', 'dict_classes', 'multi_label', 'level_save',
                          'batch_size', 'epochs', 'validation_split', 'patience', 'keras_params']:
            setattr(self, attribute, configs.get(attribute, getattr(self, attribute)))

        # Reload model
        self.model = load_model_keras(hdf5_path, custom_objects=self.custom_objects)

        # Save best hdf5 in new folder
        new_hdf5_path = os.path.join(self.model_dir, 'best.hdf5')
        shutil.copyfile(hdf5_path, new_hdf5_path)

        # Reload pipeline preprocessing
        with open(preprocess_pipeline_path, 'rb') as f:
            self.preprocess_pipeline = pickle.load(f)


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")
